#include <linux/init.h>
#include <linux/module.h>
#include <asm/io.h>
#include <mach/platform.h>
#include <linux/timer.h>
#include <linux/device.h>
#include <linux/err.h>

MODULE_LICENSE("GPL");




//데이터시트에 나온 대로 계산해서 넣은 부분 
struct GpioRegisters
{
	uint32_t GPFSEL[6];
	uint32_t Reserved1;
	uint32_t GPSET[2];
	uint32_t Reserved2;
	uint32_t GPCLR[2];
};

struct GpioRegisters *s_pGpioRegisters;


static struct timer_list s_BlinkTimer;
static int s_BlinkPeriod = 1000;


static void BlinkTimerHandler(unsigned long unused)
{
	static bool on = false;
	on = !on;


	//GPIO 18을 깜박이게 하려면 다음처럼 19번째의 값을 1로 두면됨 
	if (on) //                             19번째를 1로 세팅
		s_pGpioRegisters->GPSET[0] =  0b1000000000000000000; //켜짐
	else
		s_pGpioRegisters->GPCLR[0] =  0b1000000000000000000; //꺼짐

	//타이머 만료시간을 연장한다 
	mod_timer(&s_BlinkTimer, jiffies + msecs_to_jiffies(s_BlinkPeriod));
}

static ssize_t set_period_callback(struct device* dev,
								   struct device_attribute* attr,
								   const char* buf,
								   size_t count)
{
	long period_value = 0;

	//문자열buf를 long타입 정수period_value로 변환한다. 
	if (kstrtol(buf, 10, &period_value) < 0)
		return -EINVAL;
	if (period_value < 10)	//Safety check
		return -EINVAL;

	s_BlinkPeriod = period_value;
	return count;
}

//sysfs entry를 선언한다.
static DEVICE_ATTR(period, S_IWUSR | S_IWGRP | S_IWOTH, NULL, set_period_callback);

static struct class *s_pDeviceClass;
static struct device *s_pDeviceObject;

static int __init LedBlinkModule_init(void)
{
	int result;

	//데이터시트를 보면 GPIO_BASE는 0x 2020 0000값으로 선언해야 한다고 나와있음 
	s_pGpioRegisters = (struct GpioRegisters *)__io_address(0x20200000);
     
     //18번핀만 Output으로 설정
     //                                18-17-16-15-14-13-12-11-10-
     s_pGpioRegisters->GPFSEL[1] =0b001000000000000000000000000;

	
	//타이머를 초기화하고 콜백함수를 지정한다. 
	setup_timer(&s_BlinkTimer, BlinkTimerHandler, 0);

	
	//현재 시간(jiffies)를 기준으로 타이머 만료시간을  1000 밀리 초 만큼 연장한다.
	//jiffies변수는 시스템이 부팅된 후, 발생한 타이머 인터럽트의 수를 저장하는 변수이다. 
	result = mod_timer(&s_BlinkTimer, jiffies + msecs_to_jiffies(s_BlinkPeriod));
	BUG_ON(result < 0);



	//디바이스 클래스를 생성한다. 
	///sys/class아래에 LedBlink 디렉토리가 생성됨 
	s_pDeviceClass = class_create(THIS_MODULE, "LedBlink");
	BUG_ON(IS_ERR(s_pDeviceClass));

	
	//디바이스를 생성하고 sysfs에 등록
	//creates a device and registers it with sysfs
	///sys/class/LedBlink/LedBlink 링크됨 ... /sys/devices/virtual/LedBlink/LedBlink으로 
	///sys/devices/virtual/LedBlink/LedBlink아래에 항목 세개 생김. power  subsystem  uevent
	s_pDeviceObject = device_create(s_pDeviceClass, NULL, 0, NULL, "LedBlink");
	BUG_ON(IS_ERR(s_pDeviceObject));


	//장치의 sysfs 속성 파일을 만든다.
     ///sys/class/LedBlink/LedBlink디렉토리에 period파일이 생성됨 
	result = device_create_file(s_pDeviceObject, &dev_attr_period);
	BUG_ON(result < 0);

	return 0;
}

static void __exit LedBlinkModule_exit(void)
{
	device_remove_file(s_pDeviceObject, &dev_attr_period);
	device_destroy(s_pDeviceClass, 0);
	class_destroy(s_pDeviceClass);

	//ouput으로 바꾸었던 설정을 다시 디폴트값인 input으로 변경한다.
     s_pGpioRegisters->GPFSEL[1] =0;

	//타이머를 삭제한다.
	del_timer(&s_BlinkTimer);
}

module_init(LedBlinkModule_init);
module_exit(LedBlinkModule_exit);
